@@ -510,7 +510,7 @@ impl PageData {
         }
     }
 
-    fn remove_key(&mut self, ip: ItemPointer) {
+    fn remove_key(&mut self, ip: ItemPointer, leaf: bool) {
         let n_items = self.get_n_items();
         let size = self.get_size();
         let (item_offs, item_len) = self.get_item_offs_len(ip);
@@ -518,12 +518,15 @@ impl PageData {
             self.set_offs(i - 1, self.get_offs(i) + item_len);
         }
         let items_origin = PAGE_SIZE - size;
-        if n_items > 1 && ip + 1 == n_items && self.data[item_offs] == 0 {
-            // If we are removing last child of internal page with +inf key,
-            // then we should replace key of previous item with +inf
-            let prev_key_len = self.data[item_offs + item_len] as usize;
-            self.set_offs(ip - 1, item_offs + item_len + prev_key_len);
-            self.data[item_offs + item_len + prev_key_len] = 0u8; // zero length means +inf
+        if !leaf && n_items > 1 && ip + 1 == n_items {
+            // If we are removing last child of internal page then copy it's key to the previous item
+            let prev_item_offs = item_offs + item_len;
+            let key_len = self.data[item_offs] as usize;
+            let prev_key_len = self.data[prev_item_offs] as usize;
+            let new_offs = prev_item_offs + prev_key_len - key_len;
+            self.set_offs(ip - 1, new_offs);
+            self.data
+                .copy_within(item_offs..item_offs + prev_key_len + 1, new_offs);
         } else {
             self.data
                 .copy_within(items_origin..item_offs, items_origin + item_len);
@@ -783,21 +786,29 @@ impl Storage {
     // Allocate new page in storage and get buffer for it
     //
     fn new_page(&self, db: &mut RwLockWriteGuard<Database>) -> Result<PageGuard<'_>> {
-        let mut bm = self.buf_mgr.lock().unwrap();
         let free = db.meta.free;
         let buf;
+        let mut bm = self.buf_mgr.lock().unwrap();
         if free != 0 {
             buf = bm.get_buffer(free)?;
-            let page = self.pool[buf as usize].read().unwrap();
+            let mut page = self.pool[buf as usize].write().unwrap();
+            if (bm.pages[buf as usize].state & PAGE_RAW) != 0 {
+                self.file
+                    .read_exact_at(&mut page.data, free as u64 * PAGE_SIZE as u64)?;
+            }
             db.meta.free = page.get_u32(0);
+            page.data.fill(0u8);
         } else {
             // extend storage
             buf = bm.get_buffer(db.meta.size)?;
             db.meta.size += 1;
+            let mut page = self.pool[buf as usize].write().unwrap();
+            page.data.fill(0u8);
         }
         db.meta.used += 1;
         db.meta_updated = true;
         bm.modify_buffer(buf);
+
         Ok(PageGuard {
             buf,
             pid: bm.pages[buf as usize].id,
@@ -825,20 +836,20 @@ impl Storage {
                 }
             }
         } else if (bm.pages[buf as usize].state & PAGE_RAW) != 0 {
-            let mut page = self.pool[buf as usize].write().unwrap();
             if mode != AccessMode::WriteOnly {
                 // Read buffer if not in write-only mode
                 bm.pages[buf as usize].state = PAGE_BUSY;
                 drop(bm); // read page without holding lock
-                self.file
-                    .read_exact_at(&mut page.data, pid as u64 * PAGE_SIZE as u64)?;
+                {
+                    let mut page = self.pool[buf as usize].write().unwrap();
+                    self.file
+                        .read_exact_at(&mut page.data, pid as u64 * PAGE_SIZE as u64)?;
+                }
                 bm = self.buf_mgr.lock().unwrap();
                 if (bm.pages[buf as usize].state & PAGE_WAIT) != 0 {
                     // Somebody is waiting for us
                     self.busy_events[buf as usize % N_BUSY_EVENTS].notify_all();
                 }
-            } else {
-                page.data.fill(0u8); // memset
             }
             bm.pages[buf as usize].state = 0;
         }
@@ -959,7 +970,6 @@ impl Storage {
     fn rollback(&self, db: &mut RwLockWriteGuard<Database>) -> Result<()> {
         let mut bm = self.buf_mgr.lock().unwrap();
         let mut dirty = bm.dirty_pages;
-
         // Just throw away all dirty pages from buffer cache to force reloading of original pages
         while dirty != 0 {
             debug_assert!((bm.pages[dirty as usize].state & PAGE_DIRTY) != 0);
@@ -969,6 +979,7 @@ impl Storage {
             dirty = next;
         }
         bm.dirty_pages = 0;
+
         if db.meta_updated {
             // reread metadata from disk
             let mut page = self.pool[0].write().unwrap();
@@ -1258,7 +1269,7 @@ impl Storage {
             // leaf page
             if r < n && page.compare_key(r, key) == Ordering::Equal {
                 self.modify_page(pin.buf);
-                page.remove_key(r);
+                page.remove_key(r, true);
             }
         } else {
             // recurse to next level
@@ -1266,7 +1277,7 @@ impl Storage {
             let underflow = self.btree_remove(db, page.get_child(r), key, height - 1)?;
             if underflow {
                 self.modify_page(pin.buf);
-                page.remove_key(r);
+                page.remove_key(r, false);
             }
         }
         if page.get_n_items() == 0 {
@@ -1311,7 +1322,7 @@ impl Storage {
             self.modify_page(pin.buf);
             if r < n && page.compare_key(r, key) == Ordering::Equal {
                 // replace old value with new one: just remove old one and reinsert new key-value pair
-                page.remove_key(r);
+                page.remove_key(r, true);
             }
             self.btree_insert_in_page(db, &mut page, r, key, value)
         } else {
@@ -1481,16 +1492,25 @@ impl Storage {
                 Ok(true)
             } else {
                 path.curr = None;
-                path.stack.clear();
+                path.stack.pop();
                 Ok(false)
             }
         } else {
             debug_assert!(r < n);
-            debug_assert!(page.get_child(r) != 0);
-            while r < n && !self.find(page.get_child(r), path, key, height - 1)? {
+            loop {
+                debug_assert!(page.get_child(r) != 0);
+                if self.find(page.get_child(r), path, key, height - 1)? {
+                    return Ok(true);
+                }
+                path.stack.pop();
                 r += 1;
+                if r < n {
+                    path.stack.push(PagePos { pid, pos: r });
+                } else {
+                    break;
+                }
             }
-            Ok(r < n)
+            Ok(false)
         }
     }
 
